from efmlrs.util.log import *
from efmlrs.util.data import *


def find_redundant_metabolites(smatrix, inner_counter):
    """
    Calculates the reduced row echelon form of the stoichiometric matrix. Finds metabolites that are redundant due to
    conservation relations.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param int inner_counter: integer that counts iterative steps of nullspace compression
    :return: list redundant_metas: list of metabolites to be removed
    """
    print("Start reduced row echelon form calculations", str(inner_counter), ". This may take a while")
    echelon = smatrix.T.rref(simplify=True, pivots=False)
    columns = echelon.shape[1]
    rows = echelon.shape[0]
    redundant_metas = []
    j = 0
    for i in range(0, rows):
        row = echelon.row(i)
        while j < columns and row[j] != 1:
            redundant_metas.append(j)
            j += 1
        j += 1
    return redundant_metas


def remove_redundant_metabolites(smatrix, metabolites, redundant_metas):
    """
    Removes redundant metabolites from stoichiometric matrix and from list of metabolite names. Writes compression
    information to log file. (https://academic.oup.com/bioinformatics/article/24/19/2229/246674)

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list metabolites: list of metabolite names
    :param list redundant_metas: list of metabolites to be removed
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - metabolites - list of reduced metabolite names
    """
    for i in reversed(redundant_metas):
        log_delete_meta(metabolites[i])
        del (metabolites[i])
        smatrix.row_del(i)
    return smatrix, metabolites


def run(smatrix, metabolites):
    """
    Entry point for echelon compression. Iteratively finds inconsistencies due to conservation relations of metabolites
    in the reduced row echelon form of the stoichiometric matrix and removes them.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list metabolites: list of metabolite names
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - metabolites - list of reduced metabolite names
    """
    log_module()
    inner_counter = 1

    while 1:
        redundant_metas = find_redundant_metabolites(smatrix, inner_counter)
        if len(redundant_metas) == 0:
            break
        smatrix, metabolites = remove_redundant_metabolites(smatrix, metabolites, redundant_metas)
        inner_counter += 1

    return smatrix, metabolites



import efmlrs.postprocessing.decompressions.many2one as many2one
import efmlrs.postprocessing.decompressions.nullspace as nullspace
import efmlrs.postprocessing.decompressions.deadend as deadend


def find_counter(infofile):
    """
    Parses compression info and stores information on how many compression rounds were done during preprocessing
    and how many additional bounds have been applied.

    :param infofile: file automatically created during compressions, containing all information for decompressions
    :return:
        - round_counter - int number of compression rounds
        - bounds - int number of bounds
    """
    round_counter = 0
    bounds = 0
    file = open(infofile, "r")
    for line in file:
        if line.startswith("bounds"):
            bounds = line[7]
        if line.startswith("counter"):
            round_counter = line[8]
    file.close()
    return int(round_counter), int(bounds)


def build_reverse_mapping(infofile, counter):
    """
    Reads compression information form infofile and builds a mapping for deompressions which is in reversed order of
    the previously applied compressions.

    :param infofile: file automatically created during compressions, containing all information for decompressions
    :param int counter: number of compression rounds
    :return: mappings - list of different tuple with compression information for each compression step
    """
    mappings = []
    for i in reversed(range(1, counter + 1)):
        file = open(infofile, "r")
        tmp = []
        DE = False
        O2M = False
        NS = False

        for line in file:
            if DE is True and O2M is True and NS is True:
                break
            if line.startswith("deadend_" + str(i)):
                DE = True
                deadend_cmps = deadend.parse_info(file)
                if len(deadend_cmps) != 0:
                    for reactions in deadend_cmps:
                        tmp.append(("deadend", reactions))

            if line.startswith("many2one_" + str(i)):
                O2M = True
                iterations, post, pre = many2one.parse_info(file)
                if post != pre:
                    rea_mapping = many2one.build_merge_mapping(iterations, post)
                    tmp.append(("m2o", (rea_mapping, iterations, post)))

            if line.startswith("nullspace_" + str(i)):
                NS = True
                null_cmps = nullspace.parse_info(file)
                if len(null_cmps) != 0:
                    for infos, rea_comp, rea_uncomp in reversed(null_cmps):
                        rea_mapping = nullspace.build_mapping(infos, rea_uncomp)
                        tmp.append(("nullspace", rea_mapping))

        for element in reversed(tmp):
            mappings.append(element)

        file.close()
    return mappings


def normalize_efms(decompressed, bound_info):
    """
    Only called if model had additional boundaries. Removes lambda vector entry and additional boundary reactions from
    current efm and normalizes it.

    :param list decompressed: current efm as list
    :param int bound_info: number of additional bounds
    :return: decompressed: current efm as list
    """
    lambda_val = decompressed[-1]
    del decompressed[-(bound_info + 1):]
    if lambda_val != 1 and lambda_val != 0:
        new_compressed = [val / lambda_val for val in decompressed]
        return new_compressed
    else:
        return decompressed


def write_decompressed_efms(decompressed, outputfile):
    """
    Writes final decompressed efm to user specified output file.

    :param decompressed: current efm as list
    :param outputfile: user specified filed to write decompressed files in
    :return: None
    """
    for val in decompressed:
        val = float(val)
        outputfile.write(str(val) + " ")
    outputfile.write("\n")


def decompressing(compressed_efms, outputfile, mappings, bound_info):
    """
    Iteratively decompresses one efm after another by applying decompressions according to decompression information
    stored in mappings. If model had additional boundaries: removes lambda vector entry and additional boundary
    reactions from current efm and normalizes it. Writes final decompressed efm to user specified output file.

    :param list compressed_efms: list of lists containing compressed efms
    :param outputfile: user specified filed to write decompressed files in
    :param mappings: list of different tuples with compression information for each compression step
    :param int bound_info: number of additional bounds
    :return: None
    """
    ofile = open(outputfile, "w")
    count = 0

    for cmp_efm in compressed_efms:
        decompressed = cmp_efm
        for infotype, mappinginfo in mappings:
            if infotype == "nullspace":
                decompressed = nullspace.decompressions(decompressed, mappinginfo)

            elif infotype == "m2o":
                mapping, iterations, post = mappinginfo
                decompressed = many2one.decompressions(decompressed, mapping, iterations, post)

            elif infotype == "deadend":
                decompressed = deadend.decompressions(decompressed, mappinginfo)
                continue

        if bound_info != 0:
            normalized_efms = normalize_efms(decompressed, bound_info)
            write_decompressed_efms(normalized_efms, ofile)
        else:
            write_decompressed_efms(decompressed, ofile)

        count += 1
        if count % 1000 == 0:
            print("EFMs decompressed:", count)
    ofile.close()
    print("Decompressed EFMs:", count)


def run(compressed_efms, infofile, outputfile):
    """
    Entry point for decompressing. Reads compression information form infofile and builds a mapping for deompressions
    which is in reversed order of the previously applied compressions.  Iteratively decompresses one efm after another
    by applying decompressions according to decompression information stored in mappings. If model had additional
    boundaries: removes lambda vector entry and additional boundary reactions from current efm and normalizes it.
    Writes final decompressed efm to user specified output file.

    :param infofile: file automatically created during compressions, containing all information for decompressions
    :param list compressed_efms: compressed efms
    :param compression_infos: info file created during preprocessing
    :param outputfile: path to output file
    :return: None
    """
    counter, bounds = find_counter(infofile)
    mappings = build_reverse_mapping(infofile, counter)
    print("Start decompressions")
    decompressing(compressed_efms, outputfile, mappings, bounds)



from efmlrs.util.data import *
from efmlrs.util.log import *


def rv_check(row, reversibilities):
    """
    Checks if involved reactions are reversible.

    :param row: row of stoichiometric matrix from panda dataframe
    :param list reversibilities: list of reaction reversibilities
    :return: bool
    """
    skip = False
    for i, rev in zip(range(0, len(reversibilities)), reversibilities):
        if rev == 1:
            if row[i] != 0:
                skip = True
                break
        else:
            continue
    return skip


def count_check(row):
    """
    Checks of metabolite carries only unique fluxes. This is indicated by having only positive and one negative
    coefficient respectively only negative and one positive coefficient in the row of the stoichiometric matrix.

    :param row: row of stoichiometric matrix from panda dataframe
    :return: bool
    """
    count_pos = 0
    count_neg = 0
    for reaction in row:
        if reaction == 0:
            continue
        elif reaction < 0:
            count_neg += 1
        else:
            count_pos += 1
        if count_pos > 1 and count_neg > 1:
            break
    if count_pos == 0 or count_neg == 0:
        return True
    if count_pos > 1 and count_neg > 1 or count_pos + count_neg < 2:
        return True
    return False


def get_index(row, index):
    """
    Gets indices and multiplication factors for reactions to be removed or kept.

    :param row:
    :param list index: list of reaction indices from panda data frame header
    :return: lists of tuples with reactions indices and multiplication factors
    """

    pos = [(i, abs(j)) for i, j in zip(index, row) if j > 0]
    neg = [(i, abs(j)) for i, j in zip(index, row) if j < 0]
    if len(pos) == 1:
        return pos, neg
    else:
        return neg, pos


def is_zero(row):
    """
    Checks if row contains only zeros.

    :param row: row of stoichiometric matrix from panda dataframe
    :return: bool
    """
    return all(i == 0 for i in row)


def remove_zero_rows(smatrix, metabolites):
    """
    Finds and removes rows in stoichiometric matrix that only contain zeros. Writes information on removed metabolites
    to compression log.

    :param smatrix: smatrix: panda dataframe that contains the stoichiometric matrix
    :param list metabolites: list of metabolite names
    :return: None
    """
    drops = []
    metadrops = []
    count = 0
    for index, row in smatrix.iterrows():
        if is_zero(row) is True:
            drops.append(index)
            metadrops.append(count)
        count += 1
    smatrix.drop(drops, axis=0, inplace=True)
    for i in reversed(metadrops):
        log_delete_meta(metabolites[i])
        del metabolites[i]


def merge_compress(smatrix, reversibilities, metabolites):
    """
    Iteratively searches, merges and removes irreversible reactions with unique fluxes from the stoichiometric matrix.
    Returns list of tuples of lists with merged reaction index and the corresponding multiplication factor.

    :param smatrix: panda dataframe that contains the stoichiometric matrix
    :param list reversibilities: list of reaction reversibilities
    :param list metabolites: list of metabolite names
    :return: list compressions: list of tuples with merged reaction indices and corresponding multiplication factors
    """

    inner_counter = 1
    compress = []
    while 1:
        remove = []
        for index, row in smatrix.iterrows():
            if rv_check(row, reversibilities) is True:
                continue
            if count_check(row) is True:
                continue
            remove, keep = get_index(row, smatrix.columns.values)
            compress.append((remove, keep))
            break
        if len(remove) == 0:
            break

        remove_index, remove_factor = remove[0]
        smatrix.loc[:, remove_index] /= remove_factor

        for column_index, factor in keep:
            smatrix.loc[:, column_index] /= factor
            smatrix.loc[:, column_index] += smatrix.loc[:, remove_index]

        smatrix.drop(remove_index, axis=1, inplace=True)
        remove_zero_rows(smatrix, metabolites)
        inner_counter += 1
    return compress


def cut_reactions(compressions, reversibilities, reactions):
    """
    Removes entries of removed reactions from reaction and reversibilities lists.
    Builds new merged reaction names. Writes merge information to compression log file.

    :param list compressions: list of tuples with merged reaction indices and corresponding multiplication factors
    :param list reversibilities: list of reaction reversibilities
    :param list reactions: list of reaction names
    :return:
        - reversibilities - reduced reversibilities
        - reactions - reduced list of reactions with merged reaction names
    """
    remove_entry = []
    for remove, keep in compressions:
        index = remove[0][0]
        remove_entry.append(index)
        remove_factor = remove[0][1]
        rm_rea_index = reactions[index]
        rea_names, factors = [reactions[names[0]] for names in keep], [factor[1] for factor in keep]
        result, new_names = log_merge_many(rm_rea_index, rea_names, remove_factor, factors)

        if result is False:
            if len(rea_names) == 1:
                reactions[index], reactions[keep[0][0]] = reactions[keep[0][0]], reactions[index]
                log_merge_many(rea_names[0], [rm_rea_index], remove_factor, factors, force=True)
            else:
                log_merge_many(rm_rea_index, rea_names, remove_factor, factors, force=True)

        if len(new_names) > 0:
            for i in range(0, len(keep)):
                index, factor = keep[i]
                reactions[index] = new_names[i]

    for i in reversed(sorted(remove_entry)):
        del reversibilities[i]
        del reactions[i]
    return reversibilities, reactions


def write_o2m_info(core_name, compressions, rea_pre_merge, rea_post_merge, outer_counter):
    """
    Writes compression information to *.info file for decompression during post-processing.

    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param list compressions: list of tuples with merged reaction indices and corresponding multiplication factors
    :param int rea_pre_merge: amount of reactions before many2one compression
    :param int rea_post_merge: amount of reactions after many2one compression
    :param int outer_counter: integer that counts how many iterative steps of all compressions have been performed
    :return: None
    """
    info_file_name = core_name + ".info"
    file = open(info_file_name, "a")
    file.write("many2one_" + str(outer_counter) + "\n")
    file.write("merged " + str(rea_pre_merge) + ":" + str(rea_post_merge) + " reactions\n")

    for remove, keep in compressions:
        for rea_index, factor in remove:
            file.write("remove R" + str(rea_index) + ":" + str(factor) + " ")
        for rea_index, factor in keep:
            file.write("keep R" + str(rea_index) + ":" + str(factor) + " ")
        file.write("\n")
    file.write("end\n")
    file.close()


def run(smatrix, reactions, reversibilities, metabolites, core_name, outer_counter):
    """
    Entry point for many2one compression. Iteratively searches for irreversible reactions with
    unique fluxes. The multiplication factors for these reactions is calculated and the reactions are merged. After merging
    reactions a new merged reaction is build and the original reactions are removed from the stoichiometric matrix.
    Metabolites that become zero through merges are removed.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reactions: list of reaction names
    :param list reversibilities: list of reaction reversibilities
    :param list metabolites: list of metabolite names
    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param int outer_counter: int that counts how many iterative steps with all compressions have been performed
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - reactions - list of reduced reactions names
        - reversibilities - list of reduced reaction reversibilities
        - metabolites - list of reduced metabolite names
    """
    log_module()
    smatrix = convert_matrix2df(smatrix)
    rea_pre_merge = len(smatrix.columns.values)
    compressions = merge_compress(smatrix, reversibilities, metabolites)
    rea_post_merge = len(smatrix.columns.values)
    reversibilities, reactions = cut_reactions(compressions, reversibilities, reactions)
    write_o2m_info(core_name, compressions, rea_pre_merge, rea_post_merge, outer_counter)
    smatrix = convert_df2matrix(smatrix)

    return smatrix, reactions, reversibilities, metabolites



def get_rev_info(infofile):
    """
    Reads info file and stores reaction reversibility information in a list.

    :param infofile: info file created during preprocessing
    :return: list rev_info
    """
    file = open(infofile, "r")
    rev_info = []
    for line in file:
        if line.startswith("rv:"):
            rv = line.strip().split(":")[1]
            for val in rv.replace(" ", ""):
                if int(val) == 1:
                    rev_info.append(True)
                else:
                    rev_info.append(False)
    file.close()
    return rev_info


def remove_zeros(row):
    """
    Checks if extracted efm (row) contains only zeros.

    :param list row: extracted efm
    :return: row or None
    """
    for val in row:
        if val != 0:
            return row
    return None


def parse_lrs(inputfile, reversibilities):
    """
    Parses mplrs output file, merges splitted reactions, removes rows containing only zeros and stores compressed efms
    in a list of lists.

    :param inputfile: mplrs output file
    :param list reversibilities: list of reaction reversibilities
    :return: list compressed_efms - list of lists with compressed efms
    """
    mplrs_output = open(inputfile, "r")
    compressed_efms = []
    zero_counter = 0
    found_begin = False

    for line in mplrs_output:
        if line == "":
            continue
        if line.startswith("*"):
            continue
        if line.startswith("begin"):
            found_begin = True
            continue
        if found_begin == False:
            continue
        break

    for line in mplrs_output:
        if line.startswith("end"):
            break
        val = line.split()
        row = []

        j = 0
        i = 1
        while i < len(val):
            if reversibilities[j] == 0:
                row.append(int(val[i]))
            else:
                tmp = int(val[i]) - int(val[i + 1])
                row.append(tmp)
                i += 1
            j += 1
            i += 1

        efm = remove_zeros(row)

        if efm is None:
            zero_counter += 1
        elif efm is not None:
            compressed_efms.append(efm)
            if len(compressed_efms) % 10000 == 0:
                print("EFMS extracted:" + str(len(compressed_efms)))

    mplrs_output.close()
    return compressed_efms


def get_mplrs_efms(inputfile, infofile):
    """
    Entry poitn for get_data for mplrs. Parses mplrs output file, merges splitted reactions, removes rows containing
    only zeros and stores compressed efms in a list of lists.

    :param inputfile: mplrs output file
    :param infofile: info file created during preprocessing
    :return: list compressed_efms
    """
    reversibilities = get_rev_info(infofile)
    compressed_efms = parse_lrs(inputfile, reversibilities)
    return compressed_efms


def get_efmtool_efms(inputfile):
    """
    Entry point for get_data for efmtool. Reads efmtool output file and stores compressed efms in a list of lists.

    :param inputfile: efmtool output file
    :return: list compressed_efms
    """
    ifile = open(inputfile, "r")
    compressed_efms = []
    for line in ifile:
        cmp_efm = [float(val) for val in line.strip().split()]
        compressed_efms.append(cmp_efm)
    return compressed_efms



from argparse import ArgumentParser
import io
from contextlib import redirect_stderr
import cobra
from cobra.flux_analysis import flux_variability_analysis
import sys

def read_model(input_filename):
    """
    Reads metabolic model from sbml file using the 'cobra.io.read_sbml_model' functions. Reads io string during reading
    model and catches exchange reactions added by cobrapy. These reactions are going to be removed from the model again.
    :param input_filename: sbml file with cobrapy compatible metabolic model
    :return:
        - model - cobrapy model
        - reas_added - list of reactions added by cobrapy
    """

    stream = io.StringIO()
    with redirect_stderr(stream):
        model = cobra.io.read_sbml_model(input_filename)
    console_output = stream.getvalue()

    reas_added = []
    for text in console_output.split('\n'):
        if text.startswith("Adding exchange reaction"):
            tmp = text.split()
            reas_added.append(tmp[3])

    return model, reas_added


def get_rev_amount(model):
    """
    Counts reversible reactions in model
    :param model: model from sbml
    :return: rev_cnt (float equal to number of reversible reactions)
    """
    rev_cnt = 0
    for rea in model.reactions:
        if model.reactions.get_by_id(rea.id).reversibility is True:
            rev_cnt += 1
    return rev_cnt


def model_summary(model):
    """
    Prints out important model parameters: amount of reactions, reversible reactions, metabolites and genes. Calculates
    growth rate as well using standard FBA.
    :param model: model from sbml file
    :return: None
    """
    print('------------------------------------------------')
    print(model.name)
    print('reactions:', len(model.reactions), '(including reactions added by cobrapy)')
    reversibilities = get_rev_amount(model)
    print('reversibilities: ', reversibilities)
    print('metabolites:', len(model.metabolites))
    print('genes:', len(model.genes))
    solution_fba = model.optimize()
    print('growth rate:', solution_fba.objective_value)
    print('------------------------------------------------')


def get_reactions_only_sbml(model, reas_added):
    """
    Gets reactions that are in sbml and excludes reactions added by cobrapy.
    :param model: model from sbml file
    :param reas_added: list of reactions added by cobrapy
    :return: reactions: list of reactions that are in sbml file
    """
    reactions = []
    for rea in model.reactions:
        reactions.append(rea.id)
    for rea in reas_added:
        reactions.remove(rea)
    return reactions


def get_values(rea, fva_results):
    """
    Gets important values for evaluating reactions.
    :param rea: reaction from sbml file
    :param fva_results: FVA results from cobrapy fva function
    :return:
        - min_round - FVA minimum rounded to 7th decimal
        - min_exact - FVA minimum exact result
        - max_round - FVA maximum rounded to 7th decimal
        - max_exact - FVA maximum exact result
        - diff_round - difference between max_round - min_round
        - diff_exact - difference between max_exact - min_exact
        - lb - reaction lower bound
        - ub - reaction upper bound
    """
    rea2check_fva = fva_results.loc[rea.id, :]
    min_exact = rea2check_fva.minimum
    max_exact = rea2check_fva.maximum
    min_round = round(min_exact, 10)
    min_round = min_round + 0
    max_round = round(max_exact, 10)
    max_round = max_round + 0
    diff_exact = max_exact - min_exact
    diff_round = max_round - min_round
    lb = rea.lower_bound
    ub = rea.upper_bound
    lb, ub = convert_inf(lb, ub)
    return min_round, min_exact, max_round, max_exact, diff_round, diff_exact, lb, ub


def convert_inf(lb, ub):
    """
    Converts lower and upper bounds (-inf, inf) added by cobrapy to standard float (-1000, 1000)
    :param lb: reaction lower bound
    :param ub: reaction upper bound
    :return:
        - lb - float lower bound
        - ub - float upper bound
    """
    lb_str = str(lb)
    ub_str = str(ub)
    if lb_str == '-inf':
        lb = -1000
    if ub_str == 'inf':
        ub = 1000
    return lb, ub


def check_reactions_01(reactions, fva_results, model):
    """
    First step of checking reactions incl. getting important values for evaluation
    :param reactions: list of reactions from sbml file
    :param fva_results: FVA results from cobrapy fva function
    :param model: model from sbml file
    :return: check_reas - list of reactions to check
    """
    check_reas = {}
    for rea in reactions:
        rea = model.reactions.get_by_id(rea)
        min_round, min_exact, max_round, max_exact, diff_round, diff_exact, lb, ub = get_values(rea, fva_results)
        check_reas = check_reactions_02(lb, ub, min_round, max_round, rea, diff_round, check_reas)
    return check_reas


def check_reactions_02(lb, ub, min, max, rea, diff, check_reas):
    """
    Checks if reaction bounds for reversible/irreversible reactions are correct and calls functions for further checks.
    :param lb: reaction lower bound
    :param ub: reaction upper bound
    :param min: FVA minimum rounded to 10th decimal
    :param max: FVA maximum rounded to 10th decimal
    :param rea: current reaction
    :param diff: difference between max_round - min_round
    :param check_reas: check_reas - list of reactions to check
    :return: check_reas - list of reactions to check
    """
    values = []
    if rea.reversibility:
        if lb < 0 and ub > 0:
            check_reas = check_rev(lb, ub, min, max, rea, diff, check_reas)
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    else:
        if lb >= 0 and ub >= 0:
            check_reas = check_irrev(lb, ub, min, max, rea, diff, check_reas)
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    return check_reas


def check_irrev(lb, ub, min, max, rea, diff, check_reas):
    """
    Checks if fluxes of current irreversible reaction are truly irreversible and if reaction direction is correct.
    :param lb: reaction lower bound
    :param ub: reaction upper bound
    :param min: FVA minimum rounded to 10th decimal
    :param max: FVA maximum rounded to 10th decimal
    :param rea: current reaction
    :param diff: difference between max_round - min_round
    :param check_reas: check_reas - list of reactions to check
    :return: check_reas - list of reactions to check
    """
    values = []
    diff_min_lb = round(min - lb, 6)

    if lb <= min and ub >= max:
        if diff > 0:
            pass
        elif min == max:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    else:
        if diff_min_lb >= 0:
            pass
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    return check_reas


def check_rev(lb, ub, min, max, rea, diff, check_reas):
    """
    Checks if fluxes of current reversible reaction are truly irreversible and if reaction direction is correct.
    :param lb: reaction lower bound
    :param ub: reaction upper bound
    :param min: FVA minimum rounded to 10th decimal
    :param max: FVA maximum rounded to 10th decimal
    :param rea: current reaction
    :param diff: difference between max_round - min_round
    :param check_reas: check_reas - list of reactions to check
    :return: check_reas - list of reactions to check
    """
    values = []
    if lb <= min < 0 and ub >= max and max > 0:
        if diff >= 0:
            pass
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    else:
        if min == 0 and max == 0:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
        elif min <= 0 and max <= 0:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
        else:
            values.extend([lb, ub, min, max, rea, diff])
            check_reas[rea.id] = values
    return check_reas


def rea_write(outputpath, reactions, fva_results, model):
    """
    Writes reaction information to file including lower bound (lb), upper bound (ub), FVA minimum and maximum and
    reaction reversibility
    :param outputpath: same as input path for sbml model with '_reactions.txt' as extention
    :param reactions: list of reactions that are originally in the sbml file
    :param fva_results: results of cobrapy FVA
    :param model: model from sbml file
    :return: None
    """
    file = outputpath[:-4] + '_reactions.txt'
    print('Writing results to file:',file)
    f = open(file, 'w')
    for rea in reactions:
        rea = model.reactions.get_by_id(rea)
        min_round, min_exact, max_round, max_exact, diff_round, diff_exact, lb, ub = get_values(rea, fva_results)
        f.write(str(rea) + '\n')
        f.write('lb: ' + str(lb) + ' ub: ' + str(ub) + '\n')
        f.write('min: ' + str(min_exact) + ' max: ' + str(max_exact) + ' diff: ' + str(diff_exact) + '\n')
        f.write('reversibility: ' + str(rea.reversibility) + '\n')
        f.write('------------------------------------------------\n')
    f.close()


def main(inputsbml, fraction_optimum):
    """
    Main function that parses sbml file, performs FVA, check reaction bounds and flux directions and prints out list of
    reactions that may need correction
    :param inputsbml: sbml file of metabolic model
    :param fraction_optimum: float that defines the fraction of optimum used for cobrapy FVA
    :return: None
    """
    cobra.io.read_sbml_model(inputsbml)
    model, reas_added = read_model(inputsbml)


    print('------------------------------------------------')
    print('COBRAPY added',len(reas_added),'exchange reaction')
    model_summary(model)
    print('fraction of optimum:', fraction_optimum)
    print('------------------------------------------------')

    print('Calculating FVA...')
    fva_results = flux_variability_analysis(model, model.reactions, fraction_of_optimum=fraction_optimum)
    reactions = get_reactions_only_sbml(model, reas_added)
    print('Checking reactions...')
    check_reas = check_reactions_01(reactions, fva_results, model)
    rea_write(inputsbml, reactions, fva_results, model)

    if bool(check_reas):
        print('------------------------------------------------')
        print('Please check the following reactions:')
        print('NOTE: FVA results are rounded to 10th decimal, exact values can be found in *_reactions.txt')
        print('------------------------------------------------')
        for key, vals in check_reas.items():
            tmp = vals
            print(tmp[4])
            print('lower bound:', tmp[0], 'upper bound:', tmp[1])
            print('FVA minimum:', tmp[2], 'FVA maximum:', tmp[3], 'difference:', tmp[5])
            print('reversibility:', tmp[4].reversibility)
            print('------------------------------------------------')
        print('FVA results are rounded to 10th decimal, exact values can be found in *_reactions.txt')
    else:
        print('------------------------------------------------')
        print('All reactions seem to be fine')
        print('Exact FVA results can be found in *_reactions.txt')


if __name__ == '__main__':
    DEFFILE = '../tests/example_models/ecoli5010.xml'
    DEFFRO = 0.1

    parser = ArgumentParser(description='Quickcheck for sbml file before compressions with EFMlrs and later EFM/V calculations with efmtool or mplrs', formatter_class=SmartFormatter)
    parser.add_argument('-s', '--sbmlinput', help='input is name of the sbml model')
    parser.add_argument('-f', '--fraction_optimum', help='float, optional fraction of optimum parameter for cobrapy function: flux_variability_analysis():\n'
                                        'Must be <= 1.0. Requires that the objective value is at least the fraction times maximum objective value. \n'
                                        'A value of 0.85 for instance means that the objective has to be at least at 85% percent of its maximum \n'
                                        'https://cobrapy.readthedocs.io/en/latest/_modules/cobra/flux_analysis/variability.html', default=DEFFRO)

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    try:
        main(args.sbmlinput, args.fraction_optimum)
    except Exception:
        print('crashed...')
        raise



import cobra
import io
from contextlib import redirect_stderr
import preprocessing.boundaries as boundaries
from util.data import *

def read_model(input_filename):
    """
    Reads metabolic model from sbml file using the 'cobra.io.read_sbml_model' functions. Reads io string during reading
    model and catches exchange reactions added by cobrapy. These reactions are going to be removed from the model again.
    :param input_filename: sbml file with cobrapy compatible metabolic model
    :return:
        - model - cobrapy model
        - reas_added - list of reactions added by cobrapy
    """

    stream = io.StringIO()
    with redirect_stderr(stream):
        model = cobra.io.read_sbml_model(input_filename)
    console_output = stream.getvalue()

    reas_added = []
    for text in console_output.split('\n'):
        if text.startswith("Adding exchange reaction"):
            tmp = text.split()
            reas_added.append(tmp[3])

    return model, reas_added


def rm_reactions(model, rea_list):
    """
    Removes exchange reaction that were added by cobrapy.

    :param model: cobrapy model
    :param list rea_list: list of reaction names that will be removed
    :return:
        - model - altered cobrapy model
    """
    model.remove_reactions(rea_list)
    return model


def rm_metas_in_specified_compartment(comp, model):
    """
    Removes metabolites in specified compartment.

    :param str comp: comparment to ignore
    :param model: cobrapy model
    :return: model: altered cobrapy model
    """
    metas_in_compartment2remove = []
    for meta in model.metabolites:
        if meta.compartment == comp:
            metas_in_compartment2remove.append(meta)
    model.remove_metabolites(metas_in_compartment2remove)
    return model


def compartments_2_rm(model, comp_list: list):
    """
    Checks if compartment were specified to ignore and removes metabolites that belong to specified compartments.

    :param model: cobrapy model
    :param list comp_list: list of compartments that will be removed
    :return: model: altered cobrapy model
    """

    if len(comp_list) == 0:
        print("Ignoring compartments: None")
        return model
    else:
        print("Ignoring compartments:", comp_list)
        for comp in comp_list:
            rm_metas_in_specified_compartment(comp, model)
        return model


def orphaned_metas_rm(model):
    """
    Searches and removes orphaned metabolites from cobrapy model. Orphaned metabolites are not involved in any reaction.

    :param model: cobrapy model
    :return: model: altered cobrapy model
    """
    metabolites = [meta.id for meta in model.metabolites]
    orphaned_metas = []
    for meta in metabolites:
        meta_id = model.metabolites.get_by_id(meta)
        involved_reas = len(meta_id.reactions)
        if involved_reas == 0:
            orphaned_metas.append(meta_id)

    orphaned_metas4print = [meta.id for meta in orphaned_metas]
    if len(orphaned_metas4print) != 0:
        print("The following metabolites are orphans (not involved in any reaction) and will be removed from the model")
        print(orphaned_metas4print)
    model.remove_metabolites(orphaned_metas)
    return model


def rm_empty_reas(model):
    reactions = [rea.id for rea in model.reactions]
    i = 0
    empty_rea = []
    for rea in reactions:
        rea = model.reactions.get_by_id(rea)
        reactants = rea.reactants
        products = rea.products
        if len(reactants) == 0 and len(products) == 0:
            i += 1
            empty_rea.append(rea)
    if len(empty_rea) != 0:
        rm_reactions(model, empty_rea)
        print(len(empty_rea), 'empty reactions removed from the model')
    else:
        print('no empty reactions')
    return model


def get_smatrix(model):
    """
    Gets stoichiometric matrix from cobrapy model using the cobrapy function
    "cobra.util.array.create_stoichiometric_matrix".

    :param model: cobrapy model
    :return: matrix: stoichiometric matrix
    """
    matrix = cobra.util.array.create_stoichiometric_matrix(model)
    return matrix


def check_bounds(model, smatrix, reactions, reversibilities, metabolites):
    """
    Calls boundaries script.

    :param model: cobrapy model
    :param smatrix: stoichiometric matrix
    :param reactions: list of reaction names
    :param reversibilities: list of reaction reversibilities
    :param metabolites: list of metabolite names
    :return:
        - smatrix - matrix stoichiometric matrix
        - reactions - list of reactions names
        - reversibilities - list of reaction reversibilities
        - metabolites - list of metabolite names
        - bound_counter - integer number of added bounds
    """
    smatrix, reactions, reversibilities, metabolites, bound_counter = boundaries.run(model, smatrix, reactions,
                                                                                     reversibilities, metabolites)
    return smatrix, reactions, reversibilities, metabolites, bound_counter


def write_bound_info(core_name, bound_counter):
    """
    Writes boundary information to info file.

    :param core_name: string that consists of path to and name of the input file excluding file extension
    :param int bound_counter: integer number of added bounds
    :return: None
    """
    info_file_name = core_name + ".info"
    file = open(info_file_name, "w")
    file.write("bounds " + str(bound_counter) + "\n")
    file.close()


def run(inputfile, ignore_compartments, boundflag):
    """
    Entry point for get_data. Takes sbml file as input. Removes exchange reactions that were added by cobrapy during
    reading. Using cobrapy the model name and properties are extracted. Removes orphaned metabolites. As specified by
    user input ignores compartments a and creates boundary reactions and metabolites. Converts stoichiometric matrix
    coefficients into fractions for precise arithmetic in later calculations and converts stoichiometric matrix from
    list of lists to sympy matrix. Creates the following files: sfile (smatrix), mfile (metabolite names),
    rfile (reaction names), rvfile (reaction reversibilities) and info file.

    :param inputfile: sbml file with cobrapy compatible metabolic model
    :param str ignore_compartments: (optional) user input as string with compartment name that will be ignored
    :param bool boundflag: (optional) user input as bool flag if boundaries from sbml file will be taken into account
    :return:
        - smatrix - sympy matrix of stoichiometric matrix
        - reactions - list of reactions names
        - reversibilities - list of reaction reversibilities
        - metabolites - list of metabolite names
        - model - cobrapy model
        - core_name - path to input file without extensions
    """

    model, reas_added = read_model(inputfile)
    model = rm_reactions(model, reas_added)
    model = compartments_2_rm(model, ignore_compartments)
    model = orphaned_metas_rm(model)
    model = rm_empty_reas(model)
    model.reactions.sort()
    model.metabolites.sort()
    smatrix = get_smatrix(model)
    reactions = [rea.id for rea in model.reactions]
    reversibilities = [rea.reversibility for rea in model.reactions]
    metabolites = [meta.id for meta in model.metabolites]
    core_name = inputfile[:-4]

    if boundflag is True:
        smatrix, reactions, reversibilities, metabolites, bound_counter = check_bounds(model, smatrix, reactions,
                                                                                       reversibilities, metabolites)
        write_bound_info(core_name, bound_counter)

        smatrix_list = list(smatrix)
        write_sfile_float(core_name, smatrix_list)
        write_initial_files_with_bounds(core_name, reactions, reversibilities, metabolites)

        smatrix_fraction = convert_float2fraction(smatrix)
        smatrix = Matrix(smatrix_fraction)

    else:
        bound_counter = 0
        write_bound_info(core_name, bound_counter)

        smatrix_list = list(smatrix)
        write_sfile_float(core_name, smatrix_list)

        write_initial_files_no_bounds(core_name, reactions, reversibilities, metabolites)

        smatrix_fraction = convert_float2fraction(smatrix)
        smatrix = Matrix(smatrix_fraction)

    return smatrix, reactions, reversibilities, metabolites, model, core_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_EFMlrs
----------------------------------

Tests for `EFMlrs` module.
"""

import unittest
import tempfile
from shutil import copyfile, rmtree

import efmlrs
from efmlrs import pre
from efmlrs import post

TEST_DIR        = ''
TEMP_DIR        = 'test_EFMlrs'

class TestEFMlrs(unittest.TestCase):

    def setUp(self):
        pass

    def test_pre_at_ecoli5010(self):
        # create temporary working directory for test
        TEST_DIR = tempfile.mkdtemp(TEMP_DIR)

        # argument definitions
        PATH  = 'tests/example_models/'
        MODEL = 'ecoli5010.xml'

        # copy model
        copyfile(PATH + MODEL, TEST_DIR + "/" + MODEL)

        # do pre test (in TEST_DIR)
        pre.start(
            TEST_DIR + "/" + MODEL,
            '*',
            False
        )

        # clean up
        rmtree(TEST_DIR)
        assert(efmlrs.__version__)

    def test_post_at_ecoli5010(self):
        # create temporary working directory for test
        TEST_DIR = tempfile.mkdtemp(TEMP_DIR)

        # argument definitions
        PATH   = 'tests/example_results/'
        EFMS   = 'ecoli5010_cmp_mplrs.efms'
        INFO   = 'ecoli5010.info'
        OUTPUT = 'ecoli5010_decomp.efms'

        # copy compressed efms file to temp directory
        copyfile(PATH + EFMS, TEST_DIR + "/" + EFMS)

        # copy info file to temp directory
        copyfile(PATH + INFO, TEST_DIR + "/" + INFO)

        # do post test (in TEST_DIR)
        post.start(
            TEST_DIR + "/" + EFMS,
            TEST_DIR + "/" + OUTPUT,
            TEST_DIR + "/" + INFO,
            False
        )

        # clean up
        rmtree(TEST_DIR)
        assert(efmlrs.__version__)

    def test_quickcheck_sbml_at_ecoli5010(self):
        # create temporary working directory for test
        TEST_DIR = tempfile.mkdtemp(TEMP_DIR)

        # argument definitions
        PATH = 'tests/example_models/'
        MODEL = 'ecoli5010.xml'

        # copy model
        copyfile(PATH + MODEL, TEST_DIR + "/" + MODEL)

        # do pre test (in TEST_DIR)
        quickcheck_sbml.main(
            TEST_DIR + "/" + MODEL,
            '0.1',
        )

        # clean up
        rmtree(TEST_DIR)
        assert (efmlrs.__version__)

    def tearDown(self):
        pass



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import sys
from efmlrs.util.log import *
from efmlrs.util.data import *
import efmlrs.preprocessing.get_data as get_data
import efmlrs.preprocessing.compressions.deadend as deadend
import efmlrs.preprocessing.compressions.many2one as many2one
import efmlrs.preprocessing.compressions.nullspace as nullspace
import efmlrs.preprocessing.compressions.echelon as echelon
import efmlrs.preprocessing.mplrs_output as mplrs_output


def main(inputsbml, ignore_compartments, boundflag):
    """
    Main script of preprocessing part of EFMlrs. Calls all other scripts. Compressions are being done iteratively as
    long as compressions can be found. Loop breaks after one loop with no changes in length of metabolites or reactions.
    Writes uncompressed and compressed output files including additional versions of the sfile with integers
    (instead of fractions) to be compatible with efmtool. Writes compressed input file for mplrs algorithm. Writes info
    file that is needed for decompressions and log file that contains all information on the applied compressions.

    :param inputsbml: path to sbml input file that contains the metabolic model
    :param ignore_compartments: list of compartments to ignore
    :param boundflag: bool flag when additional bounds should be taken into account
    :return: None
    """
    efmlrs_start_compressions()
    smatrix, reactions, reversibilities, metabolites, model, core_name = get_data.run(inputsbml, ignore_compartments,
                                                                                      boundflag)
    mplrs_output.run_uncmp(core_name)
    rev_count = reversibilities4printing(reversibilities)
    print("Uncompressed network size:", smatrix.shape[1], "reactions (", rev_count, "reversible ) and", smatrix.shape[0],
          "metabolites.")

    outer_counter = 1
    print("========================================================================")
    print("START COMPRESSIONS")

    while 1:
        start_metabolites = metabolites
        start_reactions = reactions

        print("*** Compression round:", outer_counter, "***")
        print("Start deadend compression...")
        smatrix, reactions, reversibilities, metabolites = deadend.run(smatrix, reactions, reversibilities, metabolites,
                                                                       core_name, outer_counter)
        print("Done deadend compression. Network size:", smatrix.shape[0], "metabolites and", smatrix.shape[1], "reactions (", rev_count, "reversible )")

        print("Start many2one compression...")
        smatrix, reactions, reversibilities, metabolites = many2one.run(smatrix, reactions, reversibilities,
                                                                        metabolites, core_name, outer_counter)
        print("Done many2one compression. Network size:", smatrix.shape[0], "metabolites and", smatrix.shape[1], "reactions (", rev_count, "reversible )")

        print("Start nullspace compression...")
        smatrix, reactions, reversibilities, metabolites = nullspace.run(smatrix, reactions, reversibilities,
                                                                         metabolites, core_name, outer_counter)
        print("Done nullspace compression. Network size:", smatrix.shape[0], "metabolites and", smatrix.shape[1], "reactions (", rev_count, "reversible )")

        print("Start echelon compressions...")
        smatrix, metabolites = echelon.run(smatrix, metabolites)
        print("Done echelon compression. Network size:", smatrix.shape[0], "metabolites and", smatrix.shape[1], "reactions (", rev_count, "reversible )")

        end_metabolites = metabolites
        end_reactions = reactions

        if len(start_metabolites) != len(end_metabolites) or len(start_reactions) != len(end_reactions):
            outer_counter += 1
            continue
        else:
            print("*** COMPRESSIONS DONE after:", outer_counter, "rounds ***")
            print("========================================================================")
            rev_count = reversibilities4printing(reversibilities)
            print("Compressed network size:", smatrix.shape[1], "reactions (", rev_count, "reversible ) and", smatrix.shape[0], "metabolites.")
            break

    print("Writing files")
    write_info(core_name, reversibilities, outer_counter)
    write_all(smatrix, reactions, reversibilities, metabolites, core_name + "_cmp")
    mplrs_output.run_cmp(core_name)
    write_cmp_int_matrix(core_name)
    efmlrs_finish_compressions()


def start(inputsbml, ignore_compartments, bounds):
    """
    Takes all parameters from command line, checks if parameters are okay and if everything is fine initialises log file
    and calls main function if an error occurs an exception is raised and ends the program.
    :param inputsbml: path to sbml input file that contains the metabolic model
    :param ignore_compartments: list of compartments to ignore
    :param bounds: bool flag when additional bounds should be taken into account
    :return: None
    """
    try:
        log_init(inputsbml[:-4])
        main(inputsbml, ignore_compartments, bounds)
        log_close()
    except Exception:
        print("crashed...")
        log_close()
        raise


def start_from_command_line():
    """
    Entry point for preprocessing EFMlrs. Contains all information and arguments for command line call.
    Calls start function.
    :return: None
    """
    usage = '''usage: efmlrs_pre -i <metabolic_model>.xml [--ignore_compartments <compartment name>] [--bounds]'''
    parser = ArgumentParser(prog='EFMlrs', description='Process information on metabolic model from sbml,\n'
                                                       'compress stoichiometric matrix and create all necessary files\n'
                                                       'for calculating EFMs with mplrs', epilog=usage,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("-i", "--inputsbml", help="path to input sbml file ")
    parser.add_argument('--ignore_compartments', nargs='*',
                        help="name or names of compartments that will be ignored e.g. C_e,C_b")
    parser.add_argument("--bounds", action='store_true',help="if flag --bounds is set, bounds from sbml will be taken into account")
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    start(args.inputsbml, args.ignore_compartments, args.bounds)


if __name__ == "__main__":
    start_from_command_line()



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import sys
import efmlrs.postprocessing.get_data as get_data
import efmlrs.postprocessing.decompressing as decompressing
from efmlrs.util.data import *


def main(inputfile, outputfile, infofile, efmtool):
    """
    Main script of prostprocessing part of EFMlrs. Calls all other scripts. First output files are read and compressed
    efms are stored in a list. Then the decompressing process is started iteratively in reversed order compared to the
    formerly applied compressions and each compressed efm is being decompressed one after another. The deompressed efms
    are written to the user specified output file.
    :param inputfile: output containing compressed efms either form mplrs or efmtool
    :param outputfile: file in which uncompressed efms are written to
    :param infofile: file automatically created during compressions, containing all information for decompressions
    :param efmtool: paramter that indicates whether decompressions for mplrs or efmtool is being executed
    :return: None
    """
    efmlrs_start_decompressions()
    if efmtool == True:
        print("Decompressing EFMs from EFMTOOL")
        compressed_efms = get_data.get_efmtool_efms(inputfile)
        decompressing.run(compressed_efms, infofile, outputfile)

    else:
        print("Decompressing EFMs from MPLRS")
        compressed_efms = get_data.get_mplrs_efms(inputfile, infofile)
        decompressing.run(compressed_efms, infofile, outputfile)
    efmlrs_finish_decompressions()


def start(inputfile, outputfile, infofile, efmtool):
    """
    Takes all parameters form commandline, checks if parameters are okay and calls main function if an error occurs an
    exception is raised and ends the program.
    :param inputfile: output containing compressed efms either form mplrs or efmtool
    :param outputfile: file in which uncompressed efms are written to
    :param infofile: file automatically created during compressions, containing all information for decompressions
    :param efmtool: paramter that indicates whether decompressions for mplrs or efmtool is being executed
    :return: None
    """
    try:
        main(inputfile, outputfile, infofile, efmtool)
    except Exception:
        print("crashed...")
        raise

def start_from_command_line():
    """
    Entry point for prostprocessing EFMlrs. Contains all information and arguments for command line call.
    Calls start function.
    :return: None
    """
    usage = '''usage: efmlrs_post -i <input file with compressed efms> -o <output file> -info <efmlrs info file> [--efmtool]'''
    parser = ArgumentParser(prog='EFMlrs', description='Extracts EFMs from result file of mplrs,\n'
                                                       'and decompresses EFMs from efmtool and mplrs results\n'
                                                       'that have been compressed with EFMlrs', epilog=usage,
                            formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument("-i", "--inputfile", help="name of input file that contains compressed EFMs")
    parser.add_argument("-o", "--outputfile", help="name of output file")
    parser.add_argument("-info", "--infofile", help="name of info file that was created during compressions")
    parser.add_argument("--efmtool", action='store_true', help="if parameter --efmtool is given, compressed efmtool result is expected as input")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    start(args.inputfile, args.outputfile, args.infofile, args.efmtool)

if __name__ == "__main__":
    start_from_command_line()



from efmlrs.util.log import *


def check_row(row):
    """
    Checks if rows qualifies as deadend metabolite. If row consists of only zeros and either positive
    or negative entries, it is a deadend metabolite and functions returns true.

    :param row: row of stoichiometric matrix
    :return: bool
    """
    state = 0
    for val in row:
        if state == 0:
            state = val
            continue
        if state < 0:
            if val <= 0:
                continue
            else:
                return False
        if state > 0:
            if val >= 0:
                continue
            else:
                return False
    return True


def check_reactions(smatrix, i, reversibilities):
    """
    Checks if corresponding reactions to a deadend metabolite are irreversible.
    Returns list of corresponding irreversible reactions to be removed.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param int i: index of deadend metabolite
    :param list reversibilities: list of reaction reversibilities
    :return: list of reaction names to be removed from stoichiometric matrix
    """
    rm_reactions = []
    index = 0
    for val in smatrix.row(i):
        if val != 0:
            if reversibilities[index] == 0:
                rm_reactions.append(index)
            else:
                return []
        index += 1
    return rm_reactions


def find_deadends(smatrix, reversibilities):
    """
    Checks all rows of stoichiometric matrix if they are deadend metabolites.
    Checks if corresponding reactions are irreversible.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reversibilities: list of reaction reversibilities
    :return:
        - remove_reactions - list of reaction names that will be removed
        - remove_metabolites - list of metabolite names that will be removed
    """
    remove_reactions = []
    remove_metabolites = []

    for i in range(0, smatrix.shape[0]):
        tmp = check_row(smatrix.row(i))
        if tmp == False:
            continue
        else:
            rm_reactions = check_reactions(smatrix, i, reversibilities)
            if len(rm_reactions) != 0:
                remove_metabolites.append(i)
                for item in rm_reactions:
                    remove_reactions.append(item)

    remove_reactions = sorted(set(remove_reactions))
    remove_metabolites = sorted(set(remove_metabolites))
    return remove_reactions, remove_metabolites


def write_deadend_info(core_name, outer_counter, removedReactions):
    """
    Writes compression information to *.info file for decompression during post-processing.

    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param int outer_counter: int that counts how many iterative steps with all compressions have been performed
    :param list removedReactions: list of removed reaction indices
    :return: None
    """
    info_file_name = core_name + ".info"
    file = open(info_file_name, "a")
    file.write("deadend_" + str(outer_counter) + "\n")
    for reactions in removedReactions:
        file.write("remove")
        for rea in reactions:
            file.write(" " + str(rea))
        file.write("\n")
    file.write("end\n")
    file.close()


def run(smatrix, reactions, reversibilities, metabolites, core_name, outer_counter):
    """
    Entry point for deadend compression. Finds and removes deadend metabolites and corresponding reaction that only
    contain zeros after a deadend metabolite is removed. This is done iteratively as long as deadend metabolites or
    corresponding reactions are found in the stoichiometric matrix. Writes information on removed reactions and
    metabolites to compression log.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reactions: list of reaction names
    :param list reversibilities: list of reaction reversibilities
    :param list metabolites: list of metabolite names
    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param int outer_counter: int that counts how many iterative steps with all compressions have been performed
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - reactions - list of reduced reactions names
        - reversibilities - list of reduced reaction reversibilities
        - metabolites - list of reduced metabolite names
    """

    log_module()
    inner_counter = 1
    removed_reactions = []
    while 1:
        remove_reactions, remove_metabolites = find_deadends(smatrix, reversibilities)
        if len(remove_reactions) == 0 and len(remove_metabolites) == 0:
            break

        for index in reversed(remove_reactions):
            log_delete_rea(reactions[index])
            smatrix.col_del(index)
            del reactions[index]
            del reversibilities[index]
        removed_reactions.append(remove_reactions)

        for index in reversed(remove_metabolites):
            log_delete_meta(metabolites[index])
            smatrix.row_del(index)
            del metabolites[index]

        inner_counter += 1

    write_deadend_info(core_name, outer_counter, removed_reactions)

    return smatrix, reactions, reversibilities, metabolites



from efmlrs.util.data import *
from efmlrs.util.log import *


def enumerate_rows(kernel):
    """
    Builds index for rows in kernel. Rows in kernel correspond to the reactions of the stoichiometric matrix, as does
    the index built.

    :param list kernel: kernel of stoichiometric matrix
    :return: list of tuples with list of reaction values from kernel matrix and corresponding reaction index
    """
    l = len(kernel[0])
    rows = []
    rows_index = []
    tuples = []
    for j in range(0, l):
        row = [item[j] for item in kernel]
        rows.append(row)
        rows_index.append(j)
        tuples.append(row)
    return list(zip(tuples, rows_index))


def check_multiplier(row1, row2):
    """
    Checks if rows are multiples of each other and calculates multiplication factor.

    :param list row1: first row
    :param list row2: second row
    :return: bool or multiplication factor as fraction
    """
    factor = 0
    for i, j in zip(row1, row2):
        if i == 0 and j == 0:
            continue
        if i == 0 or j == 0:
            return False
        if factor == 0:
            factor = i / j
        else:
            if i / j != factor:
                return False
    return factor


def search_multiples(kernel):
    """
    Builds index for reactions in kernel matrix and searches for linear depended reactions and their multiplication
    factor.

    :param list kernel: kernel of stoichiometric matrix
    :return: list of tuples with linear depend reactions indices and their multiplication factor
    """
    K = enumerate_rows(kernel)
    results = []
    i = 0
    for row1, index1 in K:
        j = 0
        for row2, index2 in K:
            if j <= i:
                j += 1
                continue
            factor = check_multiplier(row1, row2)
            if factor != False:
                results.append((index1, index2, factor))
            j += 1
        i += 1
    return results


def test_results(column_index, index_multiples, smatrix, reversibilities):
    """
    Checks if current reaction is marked for merge, if yes merges reactions and builds new merged reaction. Checks
    reversibilities of reactions to merge. Returns information if new reaction was build, new merged reaction or empty
    matrix, information on new merged reaction reversibility, index of reaction that can be removed from stoichiometric
    matrix after merge, multiplication factor, reversibility_flag (bool variable from or link of merged reaction
    reversibilities)

    :param int column_index: reaction index of current reaction
    :param list index_multiples: list of tuples with indices and multiplication factor of linear depend reactions
    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reversibilities: list of reaction reversibilities
    :return:
        - bool - True if no reactions were merged, False if reactions were merged
        - Matrix()/merged_rea - empty matrix or sympy matrix containing new reaction after merge
        - new_rev_info - reversibility of new merged reaction
        - index2  - int with index of second reaction
        - factor - multiplication factor (type: sympy.core.number)
        - reversibility_flag - bool from or link of reversibilities of involved reactions
    """

    for index1, index2, factor in index_multiples:
        reversibility_flag = reversibilities[index1] or reversibilities[index2]
        if index1 == column_index:
            rea1 = smatrix.col(index1)
            rea2 = smatrix.col(index2)
            tmp = rea2 / factor
            merged_rea = rea1 + tmp
            if reversibilities[index1] and reversibilities[index2]:
                new_rev_info = True
            else:
                new_rev_info = False
            return False, merged_rea, new_rev_info, index2, factor, reversibility_flag

        if index2 == column_index:
            return False, Matrix(), -1, -1, factor, reversibility_flag

    reversibility_flag = reversibilities[column_index]
    return True, Matrix(), reversibilities[column_index], -1, -1, reversibility_flag


def merge_multiples(smatrix, reversibilities, reactions, index_multiples):
    """
    Merges reactions, builds new merged reaction, removes old reactions and builds new stoichiometric matrix, new
    reversibilities and new reaction name list. Writes merge information to log file.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reversibilities: list of reaction reversibilities
    :param list reactions: list of reaction names
    :param list index_multiples: list of tuples with reaction indices
    :return:
        - smatrix_reduced - reduced stoichiometric matrix
        - reversibilities_reduced - reduced list of reaction's reversibilities
        - reactions_reduced - reduced list of merged reaction names
        - merge_info - list containing reaction indices_names and their multiplication factor
    """
    smatrix_reduced = Matrix()
    reversibilities_reduced = []
    reactions_reduced = []
    merge_info = []
    j = 0

    for column_index in range(0, smatrix.shape[1]):
        use_ori, merged_rea, new_rev_info, dropped_rea, factor, reversibility_flag = test_results(column_index,
                                                                                                  index_multiples,
                                                                                                  smatrix,
                                                                                                  reversibilities)
        if use_ori is True:
            smatrix_reduced = smatrix_reduced.col_insert(j, smatrix.col(column_index))
            reversibilities_reduced.append(new_rev_info)
            reactions_reduced.append(reactions[column_index])
            merge_info.append("R" + str(column_index))
            j += 1
        else:
            if reversibility_flag is False and factor < 0:
                log_delete(reactions[column_index], reactions[dropped_rea])
            else:
                if len(merged_rea) != 0:
                    smatrix_reduced = smatrix_reduced.col_insert(j, merged_rea)
                    reversibilities_reduced.append(new_rev_info)
                    result, newnames = log_merge(reactions[dropped_rea], reactions[column_index], factor)
                    if result is False:
                        if log_merge(reactions[column_index], reactions[dropped_rea], factor) is False:
                            log_merge(reactions[dropped_rea], reactions[column_index], factor, True)
                        else:
                            reactions[dropped_rea], reactions[column_index] = reactions[column_index], reactions[
                                dropped_rea]
                    if len(newnames) > 0:
                        reactions[column_index] = newnames[0]

                    reactions_reduced.append(reactions[column_index])
                    merge_info.append("R" + str(column_index) + "," + str() + str(dropped_rea) + ":" + str(factor))
                    j += 1

            for index in reversed(range(0, len(index_multiples))):
                i1, i2, f = index_multiples[index]
                if i1 == column_index and i2 == dropped_rea:
                    continue
                if i1 == column_index or i2 == column_index or i1 == dropped_rea or i2 == dropped_rea:
                    del index_multiples[index]
    return smatrix_reduced, reversibilities_reduced, reactions_reduced, merge_info


def get_index_zero_rows(smatrix):
    """
    Checks for rows that only contain zero values and return indices of these rows.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :return: list of indices of zero rows
    """
    index_zero_rows = []
    for index in range(0, smatrix.shape[0]):
        row = [metabolite for metabolite in smatrix.row(index) if metabolite != 0]
        if bool(row) is False:
            index_zero_rows.append(index)
    return index_zero_rows


def iterate(smatrix, metabolites, reactions, reversibilities, nullspace_infos, inner_counter):
    """
    Main function of nullspace compression. Calculates kernel matrix, searches for linear depended reactions,
    merges linear dependend reactions, removes original reactions and metabolites that contain only zeros after merging
    reactions, builds new stoichiometric matrix, new list of reaction's reversibilities, new list of reaction names
    and new list of metabolite names. (https://academic.oup.com/bioinformatics/article/24/19/2229/246674)

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param ist metabolites: list of metabolite names
    :param list reactions: list of reaction names
    :param list reversibilities: list of reaction reversibilities
    :param list nullspace_infos: list containing reaction names and their multiplication factor
    :param int inner_counter: integer that counts iterative steps of nullspace compression
    :return:
        - new_smatrix - reduced stoichiometric matrix
        - metabolites - reduced list of metabolite names
        - new_reactions - reduced list of merged reaction names
        - new_reversibilities - reduced list of reaction's reversibilities
        - nullspace_infos - list containing reaction indices_names and their multiplication factor
        - bool: indicating if nullspace compression loop is done (True) or not (False)
    """
    print("Start null space compressions round", str(inner_counter), ". This may take a while")
    K = smatrix.nullspace()
    if len(K) == 0:
        print("KERNELMATRIX is EMPTY")
        return smatrix, metabolites, reactions, reversibilities, nullspace_infos, True

    index_multiples = search_multiples(K)
    if len(index_multiples) == 0:
        return smatrix, metabolites, reactions, reversibilities, nullspace_infos, True

    new_smatrix, new_reversibilities, new_reactions, merge_info = merge_multiples(smatrix, reversibilities, reactions,
                                                                                  index_multiples)
    index_zero_rows = get_index_zero_rows(new_smatrix)

    for i in reversed(index_zero_rows):
        log_delete_meta(metabolites[i])
        del (metabolites[i])
        new_smatrix.row_del(i)

    nullspace_infos.append(merge_info)
    return new_smatrix, metabolites, new_reactions, new_reversibilities, nullspace_infos, False


def write_info(core_name, nullspace_infos, outer_counter):
    """
    Writes compression information to *.info file for decompression during post-processing.

    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param list nullspace_infos: list containing reaction names and their multiplication factor
    :param int outer_counter: int that counts how many iterative steps with all compressions have been performed
    :return: None
    """
    info_file_name = core_name + ".info"
    file = open(info_file_name, "a")
    file.write("nullspace_" + str(outer_counter) + "\n")
    reversed_info = nullspace_infos[::-1]

    for info in reversed_info:
        d = len(info)
        for i in range(0, d):
            file.write(str(info[i]) + " ")
        file.write("\n")
    file.write("end\n")
    file.close()


def run(smatrix, reactions, reversibilities, metabolites, core_name, outer_counter):
    """
    Entry point for nullspace compression. Iteratively finds and merges linear dependend reactions in the kernel matrix.
    Linear depend reactions are multiples of each others in the kernel matrix. When a linear dependencies between two
    reactions is found, their multiplication factor is calculated and these reactions get merged. Thereby a new
    reaction is build and replaces the old reactions. Metabolites that become zero through merges are removed.

    :param smatrix: sympy matrix that contains the stoichiometric matrix
    :param list reactions: list of reaction names
    :param list reversibilities: list of reaction reversibilities
    :param list metabolites: list of metabolite names
    :param str core_name: string that consists of path to and name of the input file excluding file extension
    :param int outer_counter: int that counts how many iterative steps with all compressions have been performed
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - reactions - list of reduced reactions names
        - reversibilities - list of reduced reaction reversibilities
        - metabolites - list of reduced metabolite names
    """
    log_module()
    nullspace_infos = []
    inner_counter = 1

    while 1:
        smatrix, metabolites, reactions, reversibilities, infos, last = iterate(smatrix, metabolites, reactions,
                                                                                reversibilities, nullspace_infos,
                                                                                inner_counter)
        if last is False:
            inner_counter += 1
        else:
            break

    write_info(core_name, nullspace_infos, outer_counter)
    return smatrix, reactions, reversibilities, metabolites



"""
Collection of small scripts for creating log file.
"""
import inspect
from datetime import datetime

def log_init(name):
    name += "_compression.log"
    if type(log_init.logfile) == bool:
        log_init.logfile = open(name, "w")


log_init.logfile = False
log_init.search = False


def log():
    return log_init.logfile


def log_time():
    timestamp = datetime.now()
    log().write(str(timestamp) + " ")

def log_module():
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    filename = module.__name__
    log_time()
    log().write("*** " + filename + " ***\n")


def log_close():
    if type(log()) != bool:
        log().close()


def log_add():
    return "   [+] "


def log_del():
    return "   [-] "


def log_keep():
    return "   [>] "


def log_strip(rea):
    return rea.strip('"')


def merge_reaction_names(R1_keep, R2_remove):
    return R1_keep + ":" + R2_remove


def log_merge(remove, keep, keep_factor):
    log_time()
    log().write("merging ...\n")
    log_time()
    log().write(log_del() + log_strip(remove) + " factor: " + str(keep_factor) + " \n")
    log_time()
    log().write(log_add() + log_strip(keep) + " \n")
    if log_init.search is False:
        keep = merge_reaction_names(keep, remove)
        log_time()
        log().write(log_keep() + log_strip(keep) + " \n")
        return True, [keep]
    log_time()
    log().write(log_keep() + log_strip(keep) + " \n")
    return True, []


def log_merge_many(remove, keep, remove_factor, factors, force=False):
    index = 0
    log_time()
    log().write("merging " + str(len(keep)) + " ...\n")
    log_time()
    log().write(log_del() + log_strip(remove) + " factor: " + str(remove_factor) + " \n")
    names = []

    for rea in keep:
        log_time()
        log().write(log_add() + log_strip(rea) + " factor: " + str(factors[index]) + " \n")
        if log_init.search is False:
            rea = merge_reaction_names(rea, remove)
            names.append(rea)
        log_time()
        log().write(log_keep() + log_strip(rea) + "\n")
        index += 1
    return True, names


def log_delete_meta(meta):
    log_time()
    log().write("deleting metabolite " + log_strip(meta) + " \n")


def log_delete_rea(rea):
    log_time()
    log().write("deleting ... \n")
    log_time()
    log().write(log_del() + log_strip(rea) + " \n")


def log_delete(rea1, rea2):
    log_time()
    log().write("deleting ... \n")
    log_time()
    log().write(log_del() + log_strip(rea1) + " \n")
    log_time()
    log().write(log_del() + log_strip(rea2) + " \n")


def log_merge_rea(rea):
    log_time()
    log().write("merging single \n")
    log_time()
    log().write(log_add() + log_strip(rea) + " \n")
    log_time()
    log().write(log_keep() + log_strip(rea) + " \n")



def parse_info(file):
    """
    Reads decompression information from infofile and builds decompression list for deadend compressions.

    :param file: infofile automatically created during compressions, containing all information for decompressions
    :return: cmp_info: list containing decompression information for deadend compressions
    """
    cmp_infos = []
    for line in file:
        if line.startswith("remove"):
            reactions = []
            for rea in line.split()[1:]:
                reactions.append(int(rea))
            cmp_infos.append(reactions)
        if line.startswith("end"):
            break
    return cmp_infos


def decompressions(compressed, mapping):
    """
    Decompresses the part of the current compressed efm that has been compressed during deadend compression by inserting
    zeros at the index of deadend metabolites.

    :param list compressed: list containing efm for decompression
    :param mapping: list containing decompression information for deadend decompressions
    :return: compressed: list containing current efm with decompressed deadend part
    """
    [compressed.insert(val, 0) for val in mapping]
    return compressed



from fractions import Fraction


def parse_info(file):
    """
    Reads decompression information from infofile and builds decompression list for nullspace compressions.

    :param file: infofile automatically created during compressions, containing all information for decompressions
    :return: cmp_info: list containing decompression information for deadend compressions
    """
    cmp_infos = []
    for line in file:
        if line.startswith("R"):
            infos = {}
            data = line.strip().split()
            for i in data:
                tokens = i.split(",")
                if len(tokens) <= 1:
                    continue
                index_1 = tokens[0][1:]
                index_2, factor = tokens[1].split(":")
                infos[int(index_1)] = (int(index_2), Fraction(factor))
            cmp_infos.append((infos, len(data), len(data) + len(infos)))
        if line.startswith("end"):
            break
    return cmp_infos


def build_mapping(infos, rea_uncomp):
    """
    Builds list containing decompression information for nullspace decompressions.

    :param dictionary infos: dictionary containing index of merged reaction, index of original reaction and merge factor
    :param int rea_uncomp: int number equal to the amount of original reactions
    :return: mapping: list containing decompression information for nullspace decompressions
    """
    mapping = {}

    for index_1, value in infos.items():
        index_2, factor = value
        mapping[index_2] = (index_1, Fraction(factor))

    count = 0
    for i in range(0, rea_uncomp):
        if i in mapping:
            continue
        mapping[i] = count
        count += 1

    mapping = sorted(mapping.items())
    return mapping


def decompressions(cmp_efm, mapping):
    """
    Decompresses the part of the current compressed efm that has been compressed during nullspace compression. This is
    done by splitting reactions that have been merged during nullspace into two reactions. Therefore the merged reaction
    is divided through the merge factor that has been used during nullspace compressions.

    :param list cmp_efm: list containing efm for decompression
    :param mapping: list of tuples containing decompression information for nullspace decompressions
    :return: decmp_efm list containing current efm with decompressed nullspace part
    """
    decmp_efm = []

    for i, val in mapping:
        if type(val) == int:
            decmp_efm.append(cmp_efm[val])
        else:
            index = val[0]
            factor = val[1]
            mapped_index = mapping[index][1]
            new_val = Fraction(cmp_efm[mapped_index]) / factor
            decmp_efm.append(new_val)
    return decmp_efm



"""A setuptools based setup module for EFMlrs"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-

from codecs import open
from os import path
from setuptools import setup, find_packages

import versioneer

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'HISTORY.rst'), encoding='utf-8') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # NOTE: put package requirements here
    'cobra==0.17.1',
    'numpy==1.18.1',
    'pandas==1.0.1',
    'sympy==1.5.1',
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='efmlrs',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Extracts EFMs from result file of mplrs and decompresses EFMs from efmtool and mplrs results that have been compressed with EFMlrs",
    long_description=readme + '\n\n' + history,
    author="Bianca Buchner",
    author_email='bianca.buchner@gmail.com',
    url='https://github.com/BeeAnka/EFMlrs',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    entry_points={
        'console_scripts':[
            'efmlrs_pre=efmlrs.pre:start_from_command_line',
            'efmlrs_post=efmlrs.post:start_from_command_line',
            'quickcheck_sbml=quickcheck_sbml:start_from_command_line'
            ],
        },
    include_package_data=True,
    install_requires=requirements,
    license="GPLv3",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        #'Programming Language :: Python :: 3.8',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



from efmlrs.util.data import *


def search_bounds(model):
    """
    Searches cobrapy model for additional reaction bounds.

    :param model: cobrapy model
    :return: dictionary containing information on reactions with additional bounds (reaction id, lower and upper bound)
        or returns None if no additional reaction bounds were found in the model.
    """
    bounds = {}
    for r in model.reactions:
        tmp = []
        if r.reversibility is True:
            if r.lower_bound == -1000.0 and r.upper_bound == 1000.0:
                continue
            elif str(r.lower_bound) == "-inf" and str(r.upper_bound) == "inf":
                continue
            else:
                tmp.append(r.lower_bound)
                tmp.append(r.upper_bound)
                bounds[r.id] = tmp
        else:
            if r.lower_bound == 0 and r.upper_bound == 1000.0:
                continue
            elif r.lower_bound == 0 and str(r.upper_bound) == "inf":
                continue
            else:
                tmp.append(r.lower_bound)
                tmp.append(r.upper_bound)
                bounds[r.id] = tmp

    if bool(bounds) is False:
        return False
    else:
        return bounds


def add_boundary_names(bounds_info):
    """
    Extends reaction name for reactions with additional bounds by adding "min" or "max" to lower respectively upper
    bounds.

    :param bounds_info: dictionary containing information on reaction and associated bounds
    :return: dictionary containing information on reaction bounds with extended reaction name
    """
    bounds = {}
    for rea, bound in bounds_info.items():
        if bound[0] != 0 and bound[0] != -1000:
            rea_min = rea + "_min"
            bounds[rea_min] = bound[0]
        if bound[1] != 1000:
            rea_max = rea + "_max"
            bounds[rea_max] = bound[1]
    return bounds


def format_bounds(bounds_info):
    """
    Changes bound information for additional reaction bounds stored in bounds_info from str (cobrapy format) to int.

    :param bounds_info: dictionary containing information on reaction and associated bounds
    :return: dictionary containing information on reaction bounds with extended reaction name and int as bounds
    """
    for rea, bounds in bounds_info.items():
        if str(bounds[0]) == "-inf":
            del bounds[0]
            bounds.insert(0, -1000)
        elif str(bounds[1]) == "inf":
            del bounds[1]
            bounds.insert(1, 1000)
    bounds_info = add_boundary_names(bounds_info)
    return bounds_info


def extend_smatrix_1(matrix, bounds):
    """
    Extends columns of stoichiometric matrix by number of bounds plus 1 with arrays containing only zeros.

    :param matrix: stoichiometric matrix
    :param bounds: dictionary with bound information
    :return: extended stoichiometric matrix
    """
    slack_reas_array = [0] * len(matrix)
    t_matrix = matrix.transpose()
    for i in range(len(bounds) + 1):
        t_matrix = np.vstack([t_matrix, slack_reas_array])
    matrix = t_matrix.transpose()
    return matrix


def get_bounds_index(reas, bounds):
    """
    Gets index for addtional bounds according to list of reaction names.

    :param list reas: list of reaction names
    :param bounds: dictionary with bound information
    :return: list of indicis of reaction bounds
    """
    index_bounds = []
    for rea, bound in bounds.items():
        i = 0
        for reaction in reas:
            name = rea[:-4]
            if reaction == name:
                index_bounds.append(i)
            i += 1
    return index_bounds


def build_slack_metas_01(bounds, reactions):
    """
    Builds first part of slack metabolites. Length equals length of original stoichiometric matrix. Adds 1 or -1 as
    coefficient at index of slack metabolite for upper respectively lower bound.

    :param bounds: dictionary with bound information
    :param reactions: list of reaction names
    :return: list of lists with new build slack reactions
    """
    index_bounds = get_bounds_index(reactions, bounds)
    slack_metas_1 = []
    j = 0
    for rea, bounds in bounds.items():
        tmp = []
        for i in range(len(reactions)):
            if i == index_bounds[j]:
                if rea[-3:] == "max":
                    tmp.append(1)
                else:
                    tmp.append(-1)
            else:
                tmp.append(0)
        j += 1
        slack_metas_1.append(tmp)
    return slack_metas_1


def build_slack_metas_02(bounds):
    """
    Builds entries for entries for slack reactions length equals number of additional bounds.

    :param bounds: dictionary with bound information
    :return: list slack_reas_2 of lists with entries for slack reactions
    """
    slack_metas_2 = []
    j = 0
    for rea, bound in bounds.items():
        tmp = []
        for i in range(len(bounds)):
            if i == j:
                tmp.append(1)
            else:
                tmp.append(0)
        j += 1
        slack_metas_2.append(tmp)
    return slack_metas_2


def build_slack_metas_03(bounds):
    """
    Builds last part of slack metabolite (= lambda column) with actual reaction bound as coefficient.

    :param bounds: dictionary with bound information
    :return: list slack_metas_3 of list with actual reaction bound as entry
    """
    slack_metas_3 = []
    for rea, bound in bounds.items():
        tmp = []
        if rea[-3:] == "min":
            tmp.append(bound)
        else:
            new_bound = bound * -1
            tmp.append(new_bound)
        slack_metas_3.append(tmp)
    return slack_metas_3


def building_slack_metabolites(bounds, reactions):
    """
    Builds slack metabolites. Consists of three parts. First part builds list with length of original matrix and
    coefficients according to lower or upper bound. Second part build entries for slack reactions length equals number
    of additional bounds. Third part builds entry for lambda reaction, length always 1, entry is the actual reaction
    bound.

    :param bounds: dictionary with bound information
    :param reactions: list of reaction names
    :return: list slack_metas of lists with new build slack metabolites
    """
    slack_metas_1 = build_slack_metas_01(bounds, reactions)
    slack_metas_2 = build_slack_metas_02(bounds)
    slack_metas_3 = build_slack_metas_03(bounds)
    slack_metas = []
    for i in range(len(bounds)):
        tmp = slack_metas_1[i] + slack_metas_2[i] + slack_metas_3[i]
        slack_metas.append(tmp)
    return slack_metas


def extend_smatrix_2(slack_metas, smatrix_extended):
    """
    Extends stoichiometric matrix by new build slack metabolites.

    :param list slack_metas: list of lists with new build slack metabolites
    :param smatrix_extended_1: stoichiometric matrix that has been extended by columns containing only zero
    :return: new build smatrix with slack reactions and slack metabolites
    """
    for i in range(len(slack_metas)):
        smatrix_extended = np.vstack([smatrix_extended, slack_metas[i]])
    return smatrix_extended


def build_new_smatrix(smatrix, bounds, reactions):
    """
    Extends stoichiometric matrix according to additional bounds with slack reactions and slack metabolites. Builds
    and adds slack reactions and metabolites to stoichiometric matrix. Converts matrix coefficients to fractions.

    :param smatrix: stoichiometric matrix
    :param bounds: dictionary with bound information
    :param reactions: list of reaction names
    :return: extended stoichiometric matrix
    """
    extended_smatrix_1 = extend_smatrix_1(smatrix, bounds)
    slack_metas = building_slack_metabolites(bounds, reactions)
    extended_smatrix_2 = extend_smatrix_2(slack_metas, extended_smatrix_1)
    return extended_smatrix_2


def build_new_metas(bounds, meta_ori):
    """
    Builds new list of metabolite names including slack metabolites.

    :param bounds: dictionary with bound information
    :param meta_ori: list of metabolite names
    :return: new_metas list of metabolite names including added slack metabolites
    """
    slack_metas = [("MS_" + rea) for rea, bound in bounds.items()]
    new_metas = meta_ori + slack_metas
    return new_metas


def build_new_reas(bounds, reas_ori):
    """
    Builds new list of reaction names including slack metabolites.

    :param bounds: dictionary with bound information
    :param reas_ori: list of reactions names
    :return: new_reas list reaction names including added slack reactions
    """
    slack_reas = [("RS_" + rea) for rea, bound in bounds.items()]
    rea_lambda = ["RS_lambda"]
    new_reas = reas_ori + slack_reas + rea_lambda
    return new_reas


def build_new_reversibilities(bounds, revs_ori):
    """
    Builds new list of reaction reversibilities. All added reaction are irreversible.

    :param bounds: dictionary with bound information
    :param revs_ori: list of reaction reversibilities
    :return: new_revs list of reaction reversibilities including added slack reactions
    """
    slack_revs = [(False) for i in range(len(bounds) + 1)]
    new_revs = revs_ori + slack_revs
    return new_revs


def run(model, smatrix, reactions, reversibilities, metabolites):
    """
    Entry point for boundaries. Searched for additional reaction bounds. Adds slack reaction and metabolites to the
    stoichiometric matrix, the lists of reaction names and metabolie names and the list of reaction reversibilities.

    :param model: cobrapy model
    :param smatrix: stoichiometric matrix
    :param reactions: list of reaction names
    :param reversibilities: list of reaction reversibilities
    :param metabolites: list of metabolite names
    :return:
        - smatrix - sympy matrix reduced stoichiometric matrix
        - reactions - list of reactions names
        - reversibilities - list of reaction reversibilities
        - metabolites - list of metabolite names
        - bound_counter - integer number of added bounds
    """
    print("Checking for additional bounds")
    bound_info = search_bounds(model)
    if bool(bound_info) is True:
        bounds = format_bounds(bound_info)
        bound_counter = len(bounds)
        print("Found", bound_counter, "additional bounds:")
        bounds_print(bounds)

        new_smatrix = build_new_smatrix(smatrix, bounds, reactions)
        new_metabolites = build_new_metas(bounds, metabolites)
        new_reactions = build_new_reas(bounds, reactions)
        new_reversibilities = build_new_reversibilities(bounds, reversibilities)
        return new_smatrix, new_reactions, new_reversibilities, new_metabolites, bound_counter

    else:
        print("ERROR: ADDITIONAL BOUNDS SET IN PROGRAM CALL BUT NO ADDITIONAL BOUNDS FOUND IN SBML FILE")
        print("EXITING PROGRAM")
        exit(1)



from efmlrs.util.data import *


def split_reversible_reas(smatrix, reversibilities):
    """
    Splits all reversible reactions into two single reactions. One for forward and one for backward.

    :param smatrix: list of lists of stoichiometric matrix
    :param reversibilities: list of reaction reversibilities
    :return: list of lists of reconfigured stoichiometric matrix
    """
    reconfigured_smatrix = []

    for line in smatrix:
        vtmp = []
        for i in range(0, len(line)):
            if reversibilities[i] == 1:
                vtmp.append(line[i])
                vtmp.append(-line[i])
            else:
                vtmp.append(line[i])
        reconfigured_smatrix.append(vtmp)
    return reconfigured_smatrix


def write_header(mplrs_file, core_name):
    """
    Writes header for mplrs input file.

    :param mplrs_file: path to mplrs input file
    :param core_name: path to input file without extensions
    :return: None
    """
    mplrs_file.write("* " + core_name + "\n")
    mplrs_file.write("H-representation" + "\n")


def write_smatrix(mplrs_file, reconf_smatrix):
    """
    Writes reconfigured stoichiometric matrix, unity matrix and ending to mplrs input file.

    :param mplrs_file: path to mplrs input file
    :param reconf_smatrix: list of lists of reconfigured stoichiometric matrix
    :return: None
    """
    d = len(reconf_smatrix[0])
    s = len(reconf_smatrix)
    m = s + d

    mplrs_file.write("linearity " + str(s))
    for i in range(1, s + 1):
        mplrs_file.write(" " + str(i))
    mplrs_file.write("\n")
    mplrs_file.write("begin" + "\n")
    mplrs_file.write(str(m) + " " + str((d + 1)) + " rational \n")

    for line in reconf_smatrix:
        mplrs_file.write(format(0) + " ")
        for val in line:
            mplrs_file.write(format(val) + " ")
        mplrs_file.write("\n")

    for i in range(0, d):
        mplrs_file.write(format(0) + " ")
        for j in range(0, d):
            if i == j:
                mplrs_file.write(format(1) + " ")
            else:
                mplrs_file.write(format(0) + " ")
        mplrs_file.write("\n")
    mplrs_file.write("end" + "\n")
    

def write_lrs(core_name, reconf_smatrix):
    """
    Write input file for mplrs algorithm.

    :param core_name: path to input file without extensions
    :param reconf_smatrix: list of lists of reconfigured stoichiometric matrix
    :return: None
    """
    core_name += ".ine"
    mplrs_file = open(core_name, "w")
    write_header(mplrs_file, core_name)
    write_smatrix(mplrs_file, reconf_smatrix)
    mplrs_file.close()


def write_lrs_cmp(core_name, reconf_smatrix):
    """
    Write input file for mplrs algorithm.

    :param core_name: path to input file without extensions
    :param reconf_smatrix: list of lists of reconfigured stoichiometric matrix
    :return: None
    """
    core_name += "_cmp.ine"
    mplrs_file = open(core_name, "w")
    write_header(mplrs_file, core_name)
    write_smatrix(mplrs_file, reconf_smatrix)
    mplrs_file.close()


def run_cmp(core_name):
    """
    Entry point for mplrs_output. Creates input file from sfile and rvfile suitable for mplrs algorithm.

    :param core_name: path to input file without extensions
    :return: None
    """
    smatrix = read_sfile(core_name + "_cmp")
    reversibilities = read_rvfile(core_name + "_cmp")
    reactions = read_rfile(core_name + "_cmp")

    if len(smatrix) == 0:
        print("*** NO INE FILE CREATED! ***")
        if len(reactions) > 0:
            print("*** solved due to compression ***")
            print("EFMs = ", len(reactions))
        else:
            print("*** SMATRIX EMPTY! ***")
        return
    reconfigured_smatrix = split_reversible_reas(smatrix, reversibilities)
    write_lrs_cmp(core_name, reconfigured_smatrix)

def run_uncmp(core_name):
    """
    Entry point for mplrs_output. Creates input file from sfile and rvfile suitable for mplrs algorithm.

    :param core_name: path to input file without extensions
    :return: None
    """
    smatrix = read_sfile(core_name)
    reversibilities = read_rvfile(core_name)
    reactions = read_rfile(core_name)

    if len(smatrix) == 0:
        print("*** NO INE FILE CREATED! ***")
        if len(reactions) > 0:
            print("*** solved due to compression ***")
            print("EFMs = ", len(reactions))
        else:
            print("*** SMATRIX EMPTY! ***")
        return
    reconfigured_smatrix = split_reversible_reas(smatrix, reversibilities)
    write_lrs(core_name, reconfigured_smatrix)



from fractions import Fraction


def parse_info(file):
    """
    Reads decompression information from infofile and builds decompression list for one2many compressions.

    :param file: infofile automatically created during compressions, containing all information for decompressions
    :return:
        - iterations[::-1] - list of list of tuples containing information on merged reactions and merge factor
        - post - int equal to the amount of reactions before compressions
        - pre - int equal to the amount of reactions after compressions
    """
    iterations = []
    tmp = []
    post = 0
    pre = 0
    for line in file:

        if line.startswith("merged"):
            tokens = line.strip().split()
            post, pre = tokens[1].split(":")

        if line.startswith("remove"):
            infos = []

            for i in line.strip().split():
                if i == "remove" or i == "keep":
                    continue
                tokens = i.split(":")
                if len(tokens) <= 1:
                    continue
                index = int(tokens[0][1:])
                factor = Fraction(tokens[1])
                infos.append((index, factor))
            iterations.append(infos)
        if line.startswith("end"):
            break

    tmp.append(iterations[::-1])
    tmp.append(int(post))
    tmp.append(int(pre))
    return iterations[::-1], int(post), int(pre)


def build_merge_mapping(iterations, post):
    """
    Builds dictionary containing index and merge factor for all reactions.

    :param list iterations: list of list of tuples containing information on merged reactions and merge factor
    :param int post: int equal to the amount of reactions before compressions
    :return: mapping: dictionary containing index and merge factor for all reactions
    """
    insertions = []
    for i in iterations:
        index = i[0][0]
        insertions.append(index)
    insertions = sorted(insertions)
    mapping = {}
    count = 0
    for i in range(0, post):
        if i in insertions:
            continue
        mapping[count] = i
        count += 1
    return mapping


def decompressions(cmp_efm, mapping, iterations, post):
    """
    Decompresses the part of the current compressed efm that has been compressed during one2many compression. This is
    done by splitting reactions that have been merged during nullspace into two reactions. Therefore the merged reaction
    is divided through the merge factor that has been used during one2many compressions.

    :param list cmp_efm: list containing efm for decompression
    :param dict mapping: dictionary containing index and merge factor for all reactions
    :param list iterations: list of list of tuples containing information on merged reactions and merge factor
    :param int post: int equal to the amount of reactions before compressions
    :return: decmp_efm list containing current efm with decompressed one2many part
    """
    decmp_efm = [None] * post
    for i in range(0, len(cmp_efm)):
        decmp_efm[mapping[i]] = cmp_efm[i]

    for it in iterations:
        merged_index, merged_factor = it[0]
        total = 0

        for index, factor in it[1:]:
            total += decmp_efm[index]
        decmp_efm[merged_index] = total / merged_factor

        for index, factor in it[1:]:
            decmp_efm[index] /= factor

    return decmp_efm



"""
Collection of small helper scripts.
"""
from fractions import Fraction
import numpy as np
import pandas as pd
from sympy import *
import efmlrs.util.convert_matrix as conversion


def write_all(smatrix, reactions, reversibilities, metabolites, name):
    write_sfile(name, smatrix)
    write_rvfile(name, reversibilities)
    write_init_rfile(name, reactions)
    write_init_mfile(name, metabolites)


def write_initial_files_with_bounds(core_name, reactions, reversibilities, metabolites):
    write_init_rfile(core_name, reactions)
    write_rvfile(core_name, reversibilities)
    write_init_mfile(core_name, metabolites)


def write_initial_files_no_bounds(core_name, reactions, reversibilities, metabolites):
    write_init_rfile(core_name, reactions)
    write_rvfile(core_name, reversibilities)
    write_init_mfile(core_name, metabolites)


def read_sfile_int(name):
    name += ".sfile"
    sfile = open(name, "r")
    smatrix = []
    for line in sfile:
        if line == "":
            continue
        vtmp = []
        for val in line.split():
            if val == "":
                continue
            vtmp.append(int(val))
        smatrix.append(vtmp)
    sfile.close()
    return smatrix


def read_sfile(name):
    name += ".sfile"
    sfile = open(name, "r")
    smatrix = []
    for line in sfile:
        if line == "":
            continue
        vtmp = []
        for val in line.split():
            if val == "":
                continue
            vtmp.append(Fraction(val))
        smatrix.append(vtmp)
    sfile.close()
    return smatrix


def write_sfile(name, smatrix):
    name += ".sfile"
    file = open(name, "w")
    for i in range(0, smatrix.shape[0]):
        row = smatrix.row(i)
        for j in range(0, smatrix.shape[1]):
            val = row[j]
            file.write(str(val))
            file.write(" ")
        file.write("\n")
    file.close()


def write_sfile_float(name, smatrix):
    name += ".sfile"
    file = open(name, "w")
    for row in smatrix:
        cnt = 0
        for val in row:
            cnt += 1
            if cnt < len(smatrix[0]):
                check = (val).is_integer()
                if check is True:
                    file.write(str(int(val)))
                    file.write(" ")
                if check is False:
                    file.write(str(float(val)))
                    file.write(" ")
            else:
                check = (val).is_integer()
                if check is True:
                    file.write(str(int(val)))
                if check is False:
                    file.write(str(float(val)))
        file.write("\n")
    file.close()


def convert_float2fraction(matrix):
    fr_matrix = [[Fraction(str(val)) for val in line] for line in matrix]
    return fr_matrix


def convert_fraction2float(matrix):
    float_matrix = [[float(val) for val in line] for line in matrix]
    return float_matrix


def convert_matrix2df(smatrix):
    smatrix = np.array(smatrix)
    smatrix = pd.DataFrame(smatrix)
    return smatrix


def convert_df2matrix(smatrix_df):
    smatrix = Matrix(smatrix_df)
    return smatrix


def read_mfile(name):
    name += ".mfile"
    file = open(name, "r")
    metabolism = file.read().split()
    file.close()
    return metabolism


def write_mfile(name, metabolism):
    name += ".mfile"
    file = open(name, "w")
    for i in metabolism:
        file.write(i + " ")
    file.close()


def write_init_mfile(name, metabolism):
    name += ".mfile"
    file = open(name, "w")
    for i in metabolism:
        file.write('"')
        file.write(i)
        file.write('" ')
    file.close()


def read_rfile(name):
    name += ".rfile"
    file = open(name, "r")
    reactions = file.read().split()
    file.close()
    return reactions


def write_rfile(name, reactions):
    name += ".rfile"
    file = open(name, "w")
    for i in reactions:
        file.write(i + " ")
    file.close()


def write_init_rfile(name, reactions):
    name += ".rfile"
    file = open(name, "w")
    for i in reactions:
        file.write('"')
        file.write(i)
        file.write('" ')
    file.close()


def read_rvfile(name):
    name += ".rvfile"
    file = open(name, "r")
    reversibles = file.read().split()
    file.close()
    return [bool(int(i)) for i in reversibles]


def write_rvfile(name, reversibles):
    name += ".rvfile"
    file = open(name, "w")
    for i in [str(int(j)) for j in reversibles]:
        file.write(i + " ")
    file.close()


def write_efms(name, efms):
    efm_matrix = Matrix(efms)
    file = open(name, "w")
    for i in range(0, efm_matrix.shape[0]):
        row = efm_matrix.row(i)
        for j in range(0, efm_matrix.shape[1]):
            val = row[j]
            file.write(str(val))
            file.write(" ")
        file.write("\n")
    file.close()


def write_info(core_name, reversibilities, outer_counter):
    info_file_name = core_name + ".info"
    file = open(info_file_name, "a")
    file.write("rv: ")
    for i in [str(int(j)) for j in reversibilities]:
        file.write(i + " ")
    file.write("\n")
    file.write("counter " + str(outer_counter))
    file.close()


def bounds_print(bounds):
    for name, val in bounds.items():
        if name.endswith("_max"):
            print("Upper bound:", val, "for reaction", name[:-4])
        else:
            print("Lower bound:", val, "for reaction", name[:-4])


def reversibilities4printing(reversibilities):
    i = 0
    for val in reversibilities:
        if val is True:
            i += 1
    return i


def efmlrs_start_compressions():
    print('                          ')
    print(r'           EFMlrs     __ ')
    print(r'    (\   .-.   .-.   /_")')
    print(r'     \\_//^\\_//^\\_//   ')
    print(r'      `"´   `"´   `"´    ')
    print('     start compressions   ')
    print('                          ')


def efmlrs_finish_compressions():
    print('                          ')
    print(r'                    .-.  ')
    print(r'                  /  oo  ')
    print(r'   EFMlrs         \ -,_) ')
    print(r'             _..._| \ `-<')
    print(r'        {} ." .__.\' |   ')
    print(r'       {} (         /`\  ')
    print(r'       {}(`´------´   /  ')
    print(r'          `----------´   ')
    print(r'   finished compressions ')
    print(r'                          ')


def efmlrs_start_decompressions():
    print(r'                          ')
    print(r'                    .-.  ')
    print(r'                  /  oo  ')
    print(r'   EFMlrs         \ -,_) ')
    print(r'             _..._| \ `-<')
    print(r'        {} ." .__.\' |   ')
    print(r'       {} (         /`\  ')
    print(r'       {}(`´------´   /  ')
    print(r'          `----------´   ')
    print(r'      start decompressions   ')
    print(r'                             ')


def efmlrs_finish_decompressions():
    print('                          ')
    print(r'           EFMlrs     __ ')
    print(r'    (\   .-.   .-.   /_")')
    print(r'     \\_//^\\_//^\\_//   ')
    print(r'      `"´   `"´   `"´    ')
    print('   finished decompressions')
    print('                          ')


def write_uncmp_int_matrix(core_name):
    smatrix = read_sfile(core_name)
    smatrix = Matrix(smatrix)
    write_sfile(core_name, smatrix)
    smatrix = Matrix(smatrix)
    write_sfile(core_name + "_fractions", smatrix)


def write_cmp_int_matrix(core_name):
    smatrix = read_sfile(core_name + "_cmp")
    int_smatrix = conversion.run(smatrix)
    write_sfile(core_name + "_cmp", int_smatrix)



from functools import reduce
from math import gcd
from sympy import *


def get_denominators(smatrix):
    """
    Gets greatest denominator of each row in stoichiometric matrix
    :param smatrix: list of lists of stoichiometric matrix
    :return: list of greatest denominators for each row
    """
    row_denominators = [list(set([val.denominator for val in row])) for row in smatrix]
    greatest_denominators = [reduce(lambda x, y: x * y, sublist) for sublist in row_denominators]
    return greatest_denominators


def get_int_smatrix(smatrix, greatest_denominators):
    """
    Converts stoichiometric matrix from fractions to integer by multiplying each row with greatest denominator.
    :param smatrix: list of lists of stoichiometric matrix
    :param greatest_denominators: list of greatest denominators for each row
    :return: list of lists of converted stoichiometric matrix
    """
    int_smatrix = []
    i = 0
    for row in smatrix:
        new_row = []
        for val in row:
            new_val = val * greatest_denominators[i]
            new_row.append(int(new_val))
        int_smatrix.append(new_row)
        i += 1
    return int_smatrix


def get_min_smatrix(int_smatrix):
    """
    Gets greatest divisor for each row and divides each row through greatest divisor. Converts matrix from list of
    lists to sympy matrix.
    :param int_smatrix: list of lists of stoichiometric matrix with integers as coefficients
    :return: sympy matrix of stoichiometric matrix with smallest coefficients
    """
    min_smatrix = []
    for row in int_smatrix:
        greates_divisor = reduce(gcd, row)
        if greates_divisor > 1:
            new_row = [int(val / greates_divisor) for val in row]
            min_smatrix.append(new_row)
        else:
            min_smatrix.append(row)
    min_sympy_smatrix = Matrix(min_smatrix)
    return min_sympy_smatrix


def run(smatrix):
    """
    Converts stoichiometric matrix from fractions to integer by multiplying each row with greatest denominator. For
    each row in stoichiometric matrix reduces coefficients to smallest number by dividing each row through greatest
    divisor. Converts matrix from list of lists to sympy matrix.
    :param smatrix: list of lists of stoichiometric matrix
    :return: sympy matrix of stoichiometric matrix with smallest coefficients
    """
    greatest_denominators = get_denominators(smatrix)
    int_smatrix = get_int_smatrix(smatrix, greatest_denominators)
    min_smatrix = get_min_smatrix(int_smatrix)
    return min_smatrix
